#include "prjlibs-include/standards.h"
#include <stddef.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>

#include "skalibs/include/stddjb.h"
#include "skalibs/include/ttymodes.h"
#include "cvm/include/cvm/client.h"
#include "prjlibs-include/constants.h"
#include "constants.h"

char const* PROG="cvmlogin";
#define service (PROG+3)

static int read_line(int const fd, char* buf, size_t limit,
                     size_t* const plen) {
  size_t len=0;
  while (limit!=0) {
    ssize_t const nread=read(fd, buf, 1);
    if (nread<0) return -1;
    if (nread==0) break;
    --limit;
    ++len;
    if (*buf=='\n') break;
    ++buf;
  }
  *plen=len;
  return 0;
}

static buffer* logbuf;
static char logpid[(sizeof (pid_t)*CHAR_BIT+2)/3+2]="";

static void logmsg(char const* const x0, char const* const x1) {
  if (logpid[0]=='\0')
    logpid[fmt_ulong(logpid, getpid())]=' ';
  if (buffer_puts(logbuf, service)<0) _exit(100);
  if (buffer_puts(logbuf, ": ")<0) _exit(100);
  if (buffer_puts(logbuf, logpid)<0) _exit(100);
  if (buffer_puts(logbuf, x0)<0) _exit(100);
  if (x1!=null)
    if (buffer_puts(logbuf, x1)<0) _exit(100);
  if (buffer_putsflush(logbuf, "\n")<0) _exit(100);
}

static void check(char* const buf, size_t len, size_t const max,
                  char const* const name) {
  if (len==0 || buf[len-1]!='\n')
    strerr_die3x(100, PROG, ": EOF while reading ", name);
  buf[len-1]='\0';
  if (len==max)
    strerr_die4x(100, PROG, ": ", name, " too long");
  --len;
  if (byte_chr(buf, len, '\0')!=len)
    strerr_die4x(100, PROG, ": ", name, " contains null characters");
}

static void die_alloc(void) gccattr_noreturn;
static void die_alloc(void) {
  strerr_die4sys(111, PROG, err_unable, err_alloc, ": ");
}

static char const var_cvm[]="$CVM_PLAIN";

int main(int argc, char** argv) {
  stralloc sa=GEN_ALLOC_ZERO;
  char const* x;
  typedef char buf[512];
  static size_t const bufmax=(sizeof (buf))-1;
  buf ubuf, pbuf;
  size_t ulen, plen;
  ttymodes modes, modes_noecho;
  int ret, err;
  char const* const cvm=env_get(var_cvm+1);

  logbuf=buffer_2;

  if (cvm==null)
    strerr_die4x(100, PROG, ": ", var_cvm, err_notset);
  if (ttymodes_gett(fileno_stdin, &modes)!=0)
    strerr_die4sys(100, PROG, err_unable, err_tcgetattr, ": ");
  modes_noecho=modes;
  ttymodes_echo_off(&modes_noecho);

  logmsg("tty ", env_get("TTY"));

  x=env_get("PROMPT_LOGIN");
  if (x==null) x="login: ";
  if (buffer_putsflush(buffer_1, x)<0)
    strerr_die3sys(100, PROG, err_unable, "write login prompt: ");
  if (read_line(fileno_stdin, ubuf, bufmax, &ulen)!=0)
    strerr_die3sys(100, PROG, err_unable, "read username: ");
  check(ubuf, ulen, bufmax, "username");

  logmsg("attempt ", ubuf);

  if (env_get("DENY")!=null) {
    if (stralloc_copys(&sa, "ALLOW_")==0 ||
        stralloc_cats(&sa, ubuf)==0 ||
        stralloc_0(&sa)==0)
      die_alloc();
    err=(env_get(sa.s)==null);
  } else {
    if (stralloc_copys(&sa, "DENY_")==0 ||
        stralloc_cats(&sa, ubuf)==0 ||
        stralloc_0(&sa)==0)
      die_alloc();
    err=(env_get(sa.s)!=null);
  }
  if (err) {
    x="login denied";
    logmsg(x, null);
    logbuf=buffer_1;
    logmsg(x, null);
    _exit(100);
  }

  x=env_get("PROMPT_PASSWORD");
  if (x==null) x="Password: ";
  if (buffer_putsflush(buffer_1, x)<0)
    strerr_die3sys(100, PROG, err_unable, "write password prompt: ");
  if (ttymodes_sett(fileno_stdin, &modes_noecho)!=0)
    strerr_die4sys(100, PROG, err_unable, err_tcsetattr, ": ");
  ret=read_line(fileno_stdin, pbuf, bufmax, &plen);
  err=errno;
  if (ttymodes_sett(fileno_stdin, &modes)!=0)
    strerr_die4sys(100, PROG, err_unable, err_tcsetattr, ": ");
  if (ret!=0) {
    errno=err;
    strerr_die3sys(100, PROG, err_unable, "read password: ");
  }
  if (buffer_putsflush(buffer_1, "\n")<0)
    strerr_die3sys(100, PROG, err_unable, "write newline: ");
  check(pbuf, plen, bufmax, "password");

  {
    char const* credentials[]={ null, null };
    credentials[0]=pbuf;
    ret=cvm_authenticate(cvm, ubuf, null, credentials, 0);
  }
  if (ret!=0) {
    if (ret==CVME_PERMFAIL) {
      x="login incorrect";
      logmsg(x, null);
      logbuf=buffer_1;
      logmsg(x, null);
      _exit(100);
    }
    switch (ret) {
      case CVME_GENERAL:     x=": unspecified error";        break;
      case CVME_BAD_CLIDATA: x=": invalid data from client"; break;
      case CVME_BAD_MODDATA: x=": invalid data from module"; break;
      case CVME_IO:          x=": I/O error";                break;
      default:               x=err_unknown;                  break;
    }
    strerr_die3x(111, PROG, ": error authenticating", x);
  }

  logmsg("login ok", null);

  {
    pid_t pid=fork();
    switch (pid) {
      case -1: strerr_die4sys(111, PROG, err_unable, err_fork, ": ");
      case 0: break;
      default:
        pid=wait_pid(null, pid);
        err=errno;
        logmsg("logout", null);
        if (pid>=0) _exit(0);
        errno=err;
        strerr_die4sys(100, PROG, err_unable, err_wait, ": ");
    }
  }

  if (pathexec_env(var_user+1, cvm_fact_username)==0)
    strerr_die5sys(111, PROG, err_unable, err_setvar, var_user, ": ");
  if (stralloc_copyb(&sa, "", 0)==0)
    die_alloc();
  {
    char const* c;
    if (cvm_fact_str(CVM_FACT_USERID, &c)==0)
      if (pathexec_env(var_uid+1, c)==0)
        strerr_die5sys(111, PROG, err_unable, err_setvar, var_uid, ": ");
    if (cvm_fact_str(CVM_FACT_GROUPID, &c)==0) {
      if (pathexec_env(var_gid+1, c)==0)
        strerr_die5sys(111, PROG, err_unable, err_setvar, var_gid, ": ");
      if (env_get("INCLUDE_GID")!=null)
        if (stralloc_copys(&sa, " ")==0 ||
            stralloc_cats(&sa, c)==0)
          die_alloc();
    }
    while (cvm_fact_str(CVM_FACT_SUPP_GROUPID, &c)==0)
      if (stralloc_cats(&sa, " ")==0 ||
          stralloc_cats(&sa, c)==0)
        die_alloc();
  }
  x=env_get("EXTRA_GROUPS");
  if (x!=null && stralloc_cats(&sa, x)==0)
    die_alloc();
  if (stralloc_0(&sa)==0)
    die_alloc();
  if (pathexec_env(var_groups+1, sa.s)==0)
    strerr_die5sys(111, PROG, err_unable, err_setvar, var_groups, ": ");
  if (pathexec_env(var_name+1, cvm_fact_realname)==0)
    strerr_die5sys(111, PROG, err_unable, err_setvar, var_name, ": ");
  if (pathexec_env(var_home+1, cvm_fact_directory)==0)
    strerr_die5sys(111, PROG, err_unable, err_setvar, var_home, ": ");
  if (pathexec_env(var_shell+1, cvm_fact_shell)==0)
    strerr_die5sys(111, PROG, err_unable, err_setvar, var_shell, ": ");

  if (dup2(fileno_stdout, fileno_stderr)!=fileno_stderr)
    _exit(100);

  {
    char* command[]={ (char*)"setstate", (char*)"lofgGuh", (char*)"loginshell",
                      null };
    if (argc>1) ++argv;
    else argv=command;
    pathexec((char const**)argv);
    strerr_die5sys(errstat, PROG, err_unable, err_exec, argv[0], ": ");
  }
}
